SWDEV-332811 - Clean up and extend HIP unsafe atomic add
Update HIP's unsafeAtomicAdd to:
- Compile properly even when not compiling for gfx90a
- Fall back to safe atomic add on non-gfx90a architectures
- use flat atomic add for FP64 on gfx90a, instead of dynamically
checking memory spaces.
In addition, when the compiler is passed -munsafe-fp-atomics, it
will define __AMDGCN_UNSAFE_FP_ATOMICS__. When this happens, the
compiler is requesting that the HIP headers force all HIP
atomicAdd() calls on floats or doubles to use their unsafe versions.
This patch thus causes unsafeAtomicAdd() calls when that define
is seen. This call to unsafeAtomicAdd() is also done for atomicSub(),
since that calls atomicAdd underneath. This is not done for
system-scope atomicAdd because, on gfx90a, system-scope atomic FP
add instructions would need to target fine-grained memory, which is
always unsafe.
This patch also creates safeAtomicAdd() functions for float and double.
These functions will create a standalone safe atomic, even when the
application is compiled with -munsafe-fp-atomics.
Finally, this patch adds wrappers in the Nvidia path of HIP so that
these HIP functions call through to atomicAdd there as well.
Change-Id: I8af0621d3d28ea30c9278bfeea7393d03bbdac6d
[ROCm/hipother commit: b712a51e74]
This commit is contained in:
zatwierdzone przez
Maneesh Gupta
rodzic
c010936428
commit
de06c6e4b2
@@ -26,6 +26,7 @@ THE SOFTWARE.
|
||||
#include <cuda_runtime.h>
|
||||
|
||||
#include <hip/hip_runtime_api.h>
|
||||
#include "nvidia_hip_unsafe_atomics.h"
|
||||
|
||||
#define HIP_KERNEL_NAME(...) __VA_ARGS__
|
||||
|
||||
|
||||
@@ -0,0 +1,68 @@
|
||||
/*
|
||||
Copyright (c) 2022 Advanced Micro Devices, Inc. All rights reserved.
|
||||
|
||||
Permission is hereby granted, free of charge, to any person obtaining a copy
|
||||
of this software and associated documentation files (the "Software"), to deal
|
||||
in the Software without restriction, including without limitation the rights
|
||||
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
|
||||
copies of the Software, and to permit persons to whom the Software is
|
||||
furnished to do so, subject to the following conditions:
|
||||
|
||||
The above copyright notice and this permission notice shall be included in
|
||||
all copies or substantial portions of the Software.
|
||||
|
||||
THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
|
||||
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
|
||||
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
|
||||
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
|
||||
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
|
||||
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
|
||||
THE SOFTWARE.
|
||||
*/
|
||||
|
||||
#ifndef HIP_INCLUDE_HIP_NVIDIA_DETAIL_HIP_UNSAFE_ATOMICS_H
|
||||
#define HIP_INCLUDE_HIP_NVIDIA_DETAIL_HIP_UNSAFE_ATOMICS_H
|
||||
|
||||
__device__ inline float unsafeAtomicAdd(float* addr, float value) {
|
||||
return atomicAdd(addr, value);
|
||||
}
|
||||
|
||||
__device__ inline double unsafeAtomicAdd(double* addr, double value) {
|
||||
#if __CUDA_ARCH__ < 600
|
||||
unsigned long long *addr_cast = (unsigned long long*)addr;
|
||||
unsigned long long old_val = *addr_cast;
|
||||
unsigned long long expected;
|
||||
do {
|
||||
expected = old_val;
|
||||
old_val = atomicCAS(addr_cast, expected,
|
||||
__double_as_longlong(value +
|
||||
__longlong_as_double(expected)));
|
||||
} while (__double_as_longlong(expected) != __double_as_longlong(old_val));
|
||||
return old_val;
|
||||
#else
|
||||
return atomicAdd(addr, value);
|
||||
#endif
|
||||
}
|
||||
|
||||
__device__ inline float safeAtomicAdd(float* addr, float value) {
|
||||
return atomicAdd(addr, value);
|
||||
}
|
||||
|
||||
__device__ inline double safeAtomicAdd(double* addr, double value) {
|
||||
#if __CUDA_ARCH__ < 600
|
||||
unsigned long long *addr_cast = (unsigned long long*)addr;
|
||||
unsigned long long old_val = *addr_cast;
|
||||
unsigned long long expected;
|
||||
do {
|
||||
expected = old_val;
|
||||
old_val = atomicCAS(addr_cast, expected,
|
||||
__double_as_longlong(value +
|
||||
__longlong_as_double(expected)));
|
||||
} while (__double_as_longlong(expected) != __double_as_longlong(old_val));
|
||||
return old_val;
|
||||
#else
|
||||
return atomicAdd(addr, value);
|
||||
#endif
|
||||
}
|
||||
|
||||
#endif
|
||||
Reference in New Issue
Block a user